/*
Slave driver fur Lichtorgel:
 empfngt befehle ber I2C interface, setzt die PWM Ausgnge entsprechend.
 Adresse wird im eeprom gespeichert.
 optional:
 Radom Slave Adresse generieren aus einem analog read, diese in den send Buffer legen und zum Master senden. Problem: wie delektiert man ob ein anderer Salve die gleiche Adresse hat?
 
 
 */

////////////////////
//Includes ///////
/////////////////

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/crc16.h> //crc16 calculation
#include "usiTwiSlave.h"
#include "lichtorgel.h"
#include "usiTwiSlave.h"
#include <avr/eeprom.h> 
#include <avr/pgmspace.h> //handle progmemspace

#define I2C_ADDRESS 0x11 //address: 0x11...0x15
#define RESOLUTION 13 //defines resolution of output in bits
//note: with a prescaler of 8, maximum flicker free resolution is 13bit (120Hz @ 8MHz CPU clock)
#define DDR_USI DDRA
#define PORT_USI PORTA
#define PIN_USI PINA
#define PORT_USI_SDA PA6
#define PORT_USI_SCL PA4
#define PIN_USI_SDA PINA6
#define PIN_USI_SCL PINA4
#define USI_START_COND_INT USISIF
#define USI_START_VECTOR USI_STR_vect
#define USI_OVERFLOW_VECTOR USI_OVF_vect



//////////////////
//globale variablen
//////////////////
uint16_t debug = 0; //for debugging
//uint8_t i2caddress EEMEM = I2C_ADDRESS;
uint8_t colorbuffer[5] = {
  0,0,0,0,0}; //buffer for colors from master (plus crc)
uint8_t twiwatchdog = 0;
uint8_t timer0counter = 0;
uint8_t pwmbufferready = 0;
uint8_t coloroverflowvalue[3];
uint8_t coloroverflowvalue_buffer[3];
uint8_t colorcomparevalue[3];
uint8_t colorcomparevalue_buffer[3];



//lookuptables:
//for lower values (up to 90)
uint8_t dutycyclelookup_low[91] PROGMEM= {
  0	,
  1	,
  1	,
  2	,
  3	,
  3	,
  4	,
  5	,
  6	,
  7	,
  8	,
  8	,
  9	,
  10	,
  11	,
  12	,
  13	,
  14	,
  15	,
  16	,
  17	,
  18	,
  19	,
  21	,
  22	,
  23	,
  24	,
  25	,
  27	,
  28	,
  29	,
  31	,
  32	,
  33	,
  35	,
  36	,
  38	,
  39	,
  41	,
  43	,
  44	,
  46	,
  48	,
  49	,
  51	,
  53	,
  55	,
  57	,
  59	,
  61	,
  63	,
  65	,
  67	,
  69	,
  71	,
  74	,
  76	,
  78	,
  81	,
  83	,
  86	,
  88	,
  91	,
  94	,
  97	,
  99	,
  102	,
  105	,
  108	,
  111	,
  115	,
  118	,
  121	,
  125	,
  128	,
  131	,
  135	,
  139	,
  143	,
  146	,
  150	,
  154	,
  158	,
  163	,
  167	,
  171	,
  176	,
  180	,
  185	,
  190	,
  195	
};


//for higher values (bigger than 90):
uint16_t dutycyclelookup_high[42] PROGMEM= {
  200	,
  221	,
  244	,
  269	,
  296	,
  326	,
  359	,
  394	,
  433	,
  475	,
  521	,
  572	,
  627	,
  687	,
  752	,
  823	,
  901	,
  986	,
  1079	,
  1179	,
  1290	,
  1410	,
  1541	,
  1684	,
  1840	,
  2011	,
  2196	,
  2399	,
  2620	,
  2861	,
  3124	,
  3411	,
  3725	,
  4066	,
  4439	,
  4846	,
  5289	,
  5773	,
  6301	,
  6877	,
  7505	,
  8191	
};	




///////////////////
// functions
////////////////////

void FANon() //turns NFET for Fan on
{
  PORTB |= 1; //fan FET is on PB0
}
void FANoff()
{
  PORTB &= ~(1);
}

uint8_t check_crc(void) //check crc from received data
{
  uint16_t crc = 0xffff; //initial value must be ffff!
	
    crc = _crc16_update(crc, colorbuffer[RED]);
	crc = _crc16_update(crc, colorbuffer[GREEN]);
	crc = _crc16_update(crc, colorbuffer[BLUE]);

	debug = crc;
	
	uint8_t hi = (uint8_t)(crc>>8);
	uint8_t lo = (uint8_t)(crc);

  if((hi == colorbuffer[3]) && (lo == colorbuffer[4])) return 0;
  else return 1;

}

uint16_t getinterpolatedvalue(uint8_t value)
{
  //pgm_read_word 	(

  uint16_t interpolatedvalue = 0;


  //for values bigger than 90, use high side of lookup table (every 4th value in table, fist is 91)
  if(value>90) 
  {
    if((value+1)%4 == 0) 
    {
      interpolatedvalue =  pgm_read_word(&dutycyclelookup_high[(value+1)/4-23]); 
    }

    else //reconstruct the values between the closest two point(linear interpolation)
    {
      uint16_t uppervalue =  pgm_read_word(&dutycyclelookup_high[value/4-22]);
      uint16_t lowervalue =  pgm_read_word(&dutycyclelookup_high[(value/4)-23]);
      interpolatedvalue = ((uppervalue - lowervalue)/4)*((value+1)%4)+lowervalue;

    }
  }
  //for lower values, use low part of lookup table (every value saved)
  else
  {
    interpolatedvalue =  pgm_read_byte(&dutycyclelookup_low[value]);

  }
  return interpolatedvalue;
}

void setsoftpwm() //sets compare matches and overflow values of softpwm
{

  pwmbufferready = 0; //pwm buffer is now changing


  coloroverflowvalue_buffer[0] = getinterpolatedvalue(colorbuffer[RED])>>8; 
  coloroverflowvalue_buffer[1] = getinterpolatedvalue(colorbuffer[GREEN])>>8;
  //  coloroverflowvalue_buffer[2]  = getinterpolatedvalue(colorbuffer[BLUE])>>8;
  if(colorbuffer[BLUE] == 0)
  {
    TCCR1A &= ~(1 << COM1B1); //disable output 
  }
  else
  {
    OCR1B = getinterpolatedvalue(colorbuffer[BLUE]);
    TCCR1A |= (1 << COM1B1);  //enable pwm output A1
  }

  colorcomparevalue_buffer[0] = getinterpolatedvalue(colorbuffer[RED]) - ((uint16_t)coloroverflowvalue[0]<<8);
  colorcomparevalue_buffer[1] = getinterpolatedvalue(colorbuffer[GREEN])  - ((uint16_t)coloroverflowvalue[1]<<8);
  //colorcomparevalue_buffer[2] = getinterpolatedvalue(colorbuffer[BLUE])  - ((uint16_t)coloroverflowvalue[2]<<8);

  /*
//cubic curve:
   static unsigned long factor = (unsigned long)255*(unsigned long)255*(unsigned long)255/((long)(1<<RESOLUTION) - 1);
   
   
   coloroverflowvalue_buffer[0] = ((uint32_t)colorbuffer[RED]*(uint32_t)colorbuffer[RED]*(uint32_t)colorbuffer[RED])/factor/256;  // kubisch zu x-bit overflow
   coloroverflowvalue_buffer[1] = ((uint32_t)colorbuffer[GREEN]*(uint32_t)colorbuffer[GREEN]*(uint32_t)colorbuffer[GREEN])/factor/256; 
   coloroverflowvalue_buffer[2]  = ((uint32_t)colorbuffer[BLUE]*(uint32_t)colorbuffer[BLUE]*(uint32_t)colorbuffer[BLUE])/factor/256; 
   
   
   colorcomparevalue_buffer[0] =  ((uint32_t)colorbuffer[RED]*(uint32_t)colorbuffer[RED]*(uint32_t)colorbuffer[RED])/factor - (coloroverflowvalue[0]*256);
   colorcomparevalue_buffer[1] =  ((uint32_t)colorbuffer[GREEN]*(uint32_t)colorbuffer[GREEN]*(uint32_t)colorbuffer[GREEN])/factor - (coloroverflowvalue[1]*256);
   colorcomparevalue_buffer[2] =  ((uint32_t)colorbuffer[BLUE]*(uint32_t)colorbuffer[BLUE]*(uint32_t)colorbuffer[BLUE])/factor - (coloroverflowvalue[2]*256);
   */
  pwmbufferready = 1;
}


///////////////////////////////////////////////
// MAIN 
///////////////////////////////////////////////////


int main(void) {

  DDRA = 0b10100000; //port A used for PWM (OC0B, OC1B), TWI pins are set in TWI init function
  DDRB = 0b00000101;  //output for PWM (OC0A) and FAN

  PORTA = 0; //disable all outputs as init, no pullups for inputs
  PORTB = 0;

  uint32_t counter = 0; //counts main loop for rough timing

  setsoftpwm(); //initialize softpwm to zero
  ///////////////////
  //PWM TIMER settings //////
  ////////////////////

  TCCR0A = (0 << COM0A1) | (0 << COM0A0) | (0 << COM0B1) | (0 << COM0A0)  | (1 << WGM01) | (1 << WGM00);  //phase correct pwm, 8bit
  TCCR0B = (0 << WGM02) | (0 << CS02) | (1 << CS01) | (0 << CS00);  //prescaler:8 (120Hz pwm)
  TIMSK0 = (1 << TOIE0); //timer0 overflow interrupt active


  ICR1 = 0x1FFF; //13bit counter
  TCCR1A = (0 << COM1A1) | (0 << COM1A0) | (0 << COM1B1) | (0 << COM1B0)  | (1 << WGM11) | (0 << WGM10);  //fast pwm, run to ICR1 (13 bit PWM)
  TCCR1B = (1 << WGM13) | (1 << WGM12) | (0 << CS12) | (1 << CS11) | (0 << CS10);  //prescaler:8 (120Hz pwm)
  //TIMSK1 = (1 << TOIE1); //timer1 overflow interrupt active (twi watchdog)


  //disconnect output to start in off mode

  OCR0A = 0; //start with zero dutycycle on all channels
  OCR0B = 0;
  OCR1B = 0;

  FANoff();

  sei();                  // Interrupts gloabl einschalten


  //uint8_t TWI_Address = eeprom_read_byte(&i2caddress);//lamp address 11...15
  uint8_t TWI_Address = I2C_ADDRESS;
  usiTwiSlaveInit(TWI_Address); 



  /////////MAIN LOOP///////////

  while(1)
  {

    if(usiTwiDataInReceiveBuffer())
    {

      twiwatchdog = 0; //reset watchdog on receive
	  
      colorbuffer[RED] = usiTwiReceiveByte( ); //first byte sent is red channel
      colorbuffer[GREEN] = usiTwiReceiveByte( ); //second byte sent is green channel
      colorbuffer[BLUE] = usiTwiReceiveByte( ); //third byte sent is blue channel
      colorbuffer[3] =  usiTwiReceiveByte( ); //fourth byte is CRC16 high byte
      colorbuffer[4] =  usiTwiReceiveByte( ); //fifth byte is CRC16 low byte


        if(check_crc() == 0)
      {
        //command seems valid, send ok

        //if checksum is correct, return ok
        usiTwiTransmitByte(0xff);

        /*
			uint8_t timeout = 0;			
         			while(usiTwiDataInReceiveBuffer() == 0) //wait for go or no go
         			{
         				for(uint8_t i = 0; i < 16; i++)
         				{
         					asm volatile ("nop");
         				}
         				
         			   timeout++;
         			   if(timeout > 200) {
         				
         				break;
         				}
         
         			}
         			
         			 if(usiTwiDataInReceiveBuffer())
         		 {	
         uint8_t answer = usiTwiReceiveByte();
         				if(answer == 0x55) 	updatePWM(); //on success, update pwm's
         				else
         				{
         					flushTwiBuffers();
         				}
         
         			  }
         			  */

        setsoftpwm();

      }
      else
      {
	  twiwatchdog = 0; //reset watchdog 
        //if checksum is wrong, send wrong checksum
        flushTwiBuffers(); //flush all pending data
    	usiTwiTransmitByte(0x00);
      }

    }







    if((colorbuffer[RED]+colorbuffer[GREEN]+colorbuffer[BLUE]) > FANTHRETHOLD)
    {
      FANon(); //turn cooling fan on
      counter = 0;
    }
    else  //estimated: main loop takes 8 clocks minimum, this gets true after > 10 sec = time to keep cooling
    {
      FANoff();
    }


  }

  return 0;
}





ISR(TIM0_OVF_vect) //timer2 overflow
{

  if(timer0counter==0)
  {
    //switch off all outputs
    TCCR0A &= ~(1 << COM0A1); 
    TCCR0A &= ~(1 << COM0B1); 
    //TCCR1A &= ~(1 << COM1A1);

    TCCR0B = 0; //stop timer
    //TCCR1B = 0; //stop timer

    if(pwmbufferready) //if new color is ready, copy buffer
    {
      coloroverflowvalue[0] = coloroverflowvalue_buffer[0];
      colorcomparevalue[0] = colorcomparevalue_buffer[0];
      coloroverflowvalue[1] = coloroverflowvalue_buffer[1];
      colorcomparevalue[1] = colorcomparevalue_buffer[1];
      pwmbufferready = 0;
    }

    //set pwm periods for first timer cycle if necessary
    if(coloroverflowvalue[0] == 0) OCR0A = colorcomparevalue[RED]; 
    else OCR0A = 255;
    if(coloroverflowvalue[1] == 0) OCR0B = colorcomparevalue[GREEN];
    else OCR0B = 255;
    // if(coloroverflowvalue[2] == 0) OCR1A = colorcomparevalue[BLUE];
    //else OCR1A = 255;





    //synchronize timers
    // TCNT1 = 254;  //timer1 updates OCR at TOP
    TCNT0 = 255; //timer0 updates OCR at BOTTOM

    if(colorbuffer[RED] > 0)
    {
      TCCR0A |= (1 << COM0A1);  //enable pwm output A0 
    }
    if(colorbuffer[GREEN] > 0)
    {
      TCCR0A |= (1 << COM0B1);  //enable pwm output B0 
    }

    //TCCR1A |= (1 << COM1A1);  //enable pwm output A1

      //start timers

    //TCCR1B = (0 << WGM13) | (1 << WGM12) | (0 << CS12) | (1 << CS11) | (0 << CS10);  //prescaler:8 (120Hz pwm)
    TCCR0B = (0 << WGM02) | (0 << CS02) | (1 << CS01) | (0 << CS00);  //prescaler:8 (120Hz pwm)

    while(TCNT0 > 250) //wait for timer1 to overflow
    {
      asm volatile ("nop"); //wait 
    }

    //both timers need to overflow to update the OCRnx values!

    TIFR0 = 0x01; //clear timer overflow interrupt flag

 if(twiwatchdog > 16) //no command received for 16 cycles: set to receive
    {
      // Set SDA as input
      DDR_USI &= ~( 1 << PORT_USI_SDA );

      USICR =
        // enable Start Condition Interrupt
      ( 1 << USISIE ) |
        // disable Overflow Interrupt
      ( 0 << USIOIE ) |
        // set USI in Two-wire mode, no USI Counter overflow hold
      ( 1 << USIWM1 ) | ( 0 << USIWM0 ) |
        // Shift Register Clock Source = external, positive edge
      // 4-Bit Counter Source = external, both edges
      ( 1 << USICS1 ) | ( 0 << USICS0 ) | ( 0 << USICLK ) |
        // no toggle clock-port pin
      ( 0 << USITC );

      // clear all interrupt flags and reset overflow counter

      USISR = ( 1 << USI_START_COND_INT ) | ( 1 << USIOIF ) | ( 1 << USIPF ) | ( 1 << USIDC );

      //SET_USI_TO_TWI_START_CONDITION_MODE( );
	  twiwatchdog = 0;
    }
    twiwatchdog++;


  }


  //switch off output compare unit if dutycycle is reached
  if(OCR0A == 0)   TCCR0A &= ~(1 << COM0A1); 
  if(OCR0B == 0)   TCCR0A &= ~(1 << COM0B1); 
  //if(OCR1A == 0)   TCCR1A &= ~(1 << COM1A1); 

  //set compare value to zero if this is last cycle 
  if(coloroverflowvalue[RED] == timer0counter) OCR0A = 0;  
  if(coloroverflowvalue[GREEN] == timer0counter) OCR0B = 0; 
  //if(coloroverflowvalue[BLUE] == timer0counter) OCR1A = 0; 


  timer0counter++; //increse period counter, overflows at 255->0
  if(timer0counter == (uint16_t)(1<<(RESOLUTION-8))-1) timer0counter = 0;

  //now, the next timer period is evaluated:

  //set compare values, are set to 0 in next interrupt (OCR values are only updated when timer overflows!)
  if(coloroverflowvalue[RED] == timer0counter) OCR0A = colorcomparevalue[RED] ;
  if(coloroverflowvalue[GREEN] == timer0counter) OCR0B = colorcomparevalue[GREEN] ;
  // if(coloroverflowvalue[BLUE] == timer0counter) OCR1A = colorcomparevalue[BLUE] ;
}



/*

ISR(TIM1_OVF_vect) //timer1 overflow, TWI watchdog
{
    if(twiwatchdog > 252) //no command received for 16 cycles: set to receive
    {
      // Set SDA as input
      DDR_USI &= ~( 1 << PORT_USI_SDA );

      USICR =
        // enable Start Condition Interrupt
      ( 1 << USISIE ) |
        // disable Overflow Interrupt
      ( 0 << USIOIE ) |
        // set USI in Two-wire mode, no USI Counter overflow hold
      ( 1 << USIWM1 ) | ( 0 << USIWM0 ) |
        // Shift Register Clock Source = external, positive edge
      // 4-Bit Counter Source = external, both edges
      ( 1 << USICS1 ) | ( 0 << USICS0 ) | ( 0 << USICLK ) |
        // no toggle clock-port pin
      ( 0 << USITC );

      // clear all interrupt flags and reset overflow counter

      USISR = ( 1 << USI_START_COND_INT ) | ( 1 << USIOIF ) | ( 1 << USIPF ) | ( 1 << USIDC );

      //SET_USI_TO_TWI_START_CONDITION_MODE( );
	  twiwatchdog = 0;
    }
    twiwatchdog++;
}
*/





